//
// See license.txt
//
package lv.tsi.util;

import javax.sound.midi.*;

/**
 * MIDI processor.
 *
 * @version 0.3.0 9 March 2001
 * @author  Alexey Vasilyev
 */
public class SynthNote { //implements MetaEventListener {

    private final int channel = 0;

    /** MIDI volume **/
    public int velocity = 100;

    /** Root note. 60 - C **/
    public int notes_delta = 60;

    /**
     * Time between note on and note off event in
     * milliseconds. Note that on most systems, the
     * best resolution you can expect are 10 ms.
     */
    public int duration = 5;

    /** Time stamp **/
    private long tick = 0;

    private Sequencer sequencer;
    private Synthesizer synthesizer;
    private Track track;

    /** MIDI Sequence. Is public for externally allowing
        to be saved in .mid file **/
    public Sequence sequence;

    private boolean append = false;

    /**
     * Create MIDI objects and open MIDI device.
     */
    public SynthNote() {
        open();
        //sequencer.addMetaEventListener(this);
    }

    /**
     * Open MIDI device.
     */
    public void open() {
        System.out.print("Opening MIDI... ");
        try {
            if (synthesizer == null) {
                if ((synthesizer = MidiSystem.getSynthesizer()) == null) {
                    System.out.println("getSynthesizer() failed!");
                    return;
                }
            }
            synthesizer.open();
            sequencer = MidiSystem.getSequencer();
        } catch (Exception ex) { ex.printStackTrace(); return; }
        System.out.println("ok.");
    }

    /**
     * Close MIDI device.
     */
    public void close() {
        System.out.print("Closing MIDI... ");
        if (synthesizer != null) {
            synthesizer.close();
        }
        if (sequencer != null) {
            sequencer.close();
        }
        sequencer = null;
        synthesizer = null;
        System.out.println("ok.");
    }

    /**
     * Play melody.
     *
     * @param note  MIDI key number
     */
    public synchronized void playMelody( int[] notes ) {
        System.out.print("  Melody: ");
        if (sequencer.isRunning()) {
            sequencer.stop();
        }
        createNewMelody();
        for (int i=0; i<notes.length; i++) {
             addNote( notes[i] );
        }
        startPlaying();
        if (recording)
            appendToBuf( notes );
        System.out.println();
    }


//////////////////////
// Recording section
//
    private int[]   record = new int[0];
    private boolean recording;

    private void appendToBuf( int[] notes ) {
        int[] newrec = new int[ record.length + notes.length ];
        System.arraycopy(record, 0, newrec, 0, record.length );
        System.arraycopy( notes, 0, newrec, record.length, notes.length );
        record = newrec;
    }

    public void startRecording() {
        recording = true;
        record = new int[0];
    }
    public void stopRecording() {
        recording = false;
    }
    public boolean isRecording() {
        return recording;
    }
    public void playRecorded() {
        stopRecording();
        playMelody( record );
    }
//////////////////////

    /**
     * Add note to MIDI sequence.
     *
     * @param note  MIDI key number
     */
    public void addNote(int note) {
        System.out.print(note+";");
        if (note > 11) {
            tick += duration;
            return;
        }
        note += notes_delta;
        createEvent(ShortMessage.NOTE_ON,  channel, note, tick);
        tick += duration;
        createEvent(ShortMessage.NOTE_OFF, channel, note, tick);
    }

    /**
     * Create new melody.
     */
    public void createNewMelody() {
        tick = 0;
        try {
           if (sequence==null){
               sequence = new Sequence(Sequence.PPQ, 10);
           }
        } catch (Exception ex) { ex.printStackTrace(); }

        if (track != null)
            sequence.deleteTrack(track);
        track = sequence.createTrack();
        createEvent(ShortMessage.PROGRAM_CHANGE, channel, 0, 0);
    }

    /**
     * Play note.
     *
     * @param note  MIDI key number
     */
    public void startPlaying() {
        // set and start the sequencer.
        try {
            if (!sequencer.isOpen())
                sequencer.open();
            sequencer.setSequence(sequence);
        } catch (Exception ex) { ex.printStackTrace(); }
        sequencer.start();
        //sequencer.setTempoInBPM( tempo );
    }

    private void createEvent(int type, int chan, int num, long tick) {
        ShortMessage message = new ShortMessage();
        try {
            //System.out.println( "Type: " +type+ "  Num: " + num + " Tick: " +tick );
            message.setMessage(type, channel, num, velocity);
            //message.setMessage(type, num, velocity);
            MidiEvent event = new MidiEvent( message, tick );
            track.add(event);
        } catch (Exception ex) { ex.printStackTrace(); }
    }

    /*public void meta(MetaMessage message) {
        //System.out.println( "Msg type: " +message.getType());
        if (message.getType() == 47) {  // 47 is end of track
            //sequencer.close();
            //sequencer.stop();
        }
    }*/

}
